"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'
from collections import namedtuple
from urlparse import urlparse
from rdflib import ConjunctiveGraph, URIRef, BNode, RDF, Literal
import networkx as nx

# prefixes_dict = fountain.prefixes
# prefixes = [(uri, p) for (p, uri) in prefixes_dict.items()]


def extend_uri(uri, prefixes):
    if ':' in uri:
        prefix_parts = uri.split(':')
        if len(prefix_parts) == 2 and prefix_parts[0] in prefixes:
            return prefixes[prefix_parts[0]] + prefix_parts[1]

    return uri


def is_variable(arg):
    return arg.startswith('?')


def is_uri(uri, prefixes):
    if uri.startswith('<') and uri.endswith('>'):
        uri = uri.lstrip('<').rstrip('>')
        parse = urlparse(uri, allow_fragments=True)
        return bool(len(parse.scheme))
    if ':' in uri:
        prefix_parts = uri.split(':')
        return len(prefix_parts) == 2 and prefix_parts[0] in prefixes

    return False


class TP(namedtuple('TP', "s p o")):
    @classmethod
    def _make(cls, iterable, new=tuple.__new__, len=len, prefixes=None):
        def transform_elm(elm):
            if is_variable(elm):
                return elm
            elif is_uri(elm, prefixes):
                elm = extend_uri(elm, prefixes)
                return URIRef(elm.lstrip('<').rstrip('>'))
            elif elm == 'a':
                return RDF.type
            else:
                return Literal(elm)

        if prefixes is None:
            prefixes = []
        res = filter(lambda x: x, map(transform_elm, iterable))
        if len(res) == 3:
            return new(cls, res)

        raise TypeError('Bad TP arguments: {}'.format(iterable))

    def __repr__(self):
        def elm_to_string(elm):
            if isinstance(elm, URIRef):
                if elm == RDF.type:
                    return 'a'
                return '<%s>' % elm

            return str(elm)
        strings = map(elm_to_string, [self.s, self.p, self.o])
        return '{} {} {}'.format(*strings)

    @staticmethod
    def from_string(st, prefixes):
        if st.endswith('"'):
            parts = [st[st.find('"'):]]
            st = st.replace(parts[0], '').rstrip()
            parts = st.split(" ") + parts
        else:
            parts = st.split(' ')
        return TP._make(parts, prefixes=prefixes)


class AgoraGP(object):
    def __init__(self, prefixes):
        self._tps = []
        self.__prefixes = prefixes

    @property
    def triple_patterns(self):
        return self._tps

    @property
    def prefixes(self):
        return self.__prefixes

    @property
    def graph(self):
        g = ConjunctiveGraph()
        for prefix in self.__prefixes:
            g.bind(prefix, self.__prefixes[prefix])
        variables = {}
        contexts = {}

        def nodify(elm):
            if is_variable(elm):
                if not (elm in variables):
                    elm_node = BNode(elm)
                    variables[elm] = elm_node
                return variables[elm]
            else:
                if elm == 'a':
                    return RDF.type
                elif elm.startswith('"'):
                    return Literal(elm.lstrip('"').rstrip('"'))
                else:
                    try:
                        return float(elm)
                    except ValueError:
                        return URIRef(elm)

        nxg = nx.Graph()
        for (s, p, o) in self._tps:
            nxg.add_nodes_from([s, o])
            nxg.add_edge(s, o)

        contexts = dict([(str(index), c) for (index, c) in enumerate(nx.connected_components(nxg))])

        for (s, p, o) in self._tps:
            s_node = nodify(s)
            o_node = nodify(o)
            p_node = nodify(p)

            context = None
            for uid in contexts:
                if s in contexts[uid]:
                    context = str(uid)

            g.get_context(context).add((s_node, p_node, o_node))

        return g

    @staticmethod
    def from_string(st, prefixes):
        gp = None
        if st.startswith('{') and st.endswith('}'):
            st = st.replace('{', '').replace('}', '').strip()
            tps = st.split('.')
            tps = map(lambda x: x.strip(), filter(lambda y: y != '', tps))
            gp = AgoraGP(prefixes)
            for tp in tps:
                gp.triple_patterns.append(TP.from_string(tp, gp.prefixes))
        return gp

    def __repr__(self):
        tp_strings = map(lambda x: str(x), self._tps)
        return '{ %s}' % reduce(lambda x, y: (x + '%s . ' % str(y)), tp_strings, '')






"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

from rdflib import ConjunctiveGraph, URIRef, BNode, RDF, Literal
from rdflib.namespace import Namespace, XSD, RDFS

AGORA = Namespace('http://agora.org#')


def __extend_uri(prefixes, short):
    (prefix, u) = short.split(':')
    try:
        return URIRef(prefixes[prefix] + u)
    except KeyError:
        return short


def graph_plan(plan, fountain):
    plan_graph = ConjunctiveGraph()
    plan_graph.bind('agora', AGORA)
    prefixes = plan.get('prefixes')
    ef_plan = plan.get('plan')
    tree_lengths = {}
    s_trees = set([])
    patterns = {}

    for (prefix, u) in prefixes.items():
        plan_graph.bind(prefix, u)

    def __get_pattern_node(p):
        if p not in patterns:
            patterns[p] = BNode('tp_{}'.format(len(patterns)))
        return patterns[p]

    def __inc_tree_length(tree, l):
        if tree not in tree_lengths:
            tree_lengths[tree] = 0
        tree_lengths[tree] += l

    def __add_variable(p_node, vid, subject=True):
        sub_node = BNode(str(vid).replace('?', 'var_'))
        if subject:
            plan_graph.add((p_node, AGORA.subject, sub_node))
        else:
            plan_graph.add((p_node, AGORA.object, sub_node))
        plan_graph.set((sub_node, RDF.type, AGORA.Variable))
        plan_graph.set((sub_node, RDFS.label, Literal(str(vid), datatype=XSD.string)))

    def include_path(elm, p_seeds, p_steps):
        elm_uri = __extend_uri(prefixes, elm)
        path_g = plan_graph.get_context(elm_uri)
        b_tree = BNode(elm_uri)
        s_trees.add(b_tree)
        path_g.set((b_tree, RDF.type, AGORA.SearchTree))
        path_g.set((b_tree, AGORA.fromType, elm_uri))

        for seed in p_seeds:
            path_g.add((b_tree, AGORA.hasSeed, URIRef(seed)))

        previous_node = b_tree
        __inc_tree_length(b_tree, len(p_steps))
        for j, step in enumerate(p_steps):
            prop = step.get('property')
            b_node = BNode(previous_node.n3() + prop)
            if j < len(p_steps) - 1 or pattern[1] == RDF.type:
                path_g.add((b_node, AGORA.onProperty, __extend_uri(prefixes, prop)))
            path_g.add((b_node, AGORA.expectedType, __extend_uri(prefixes, step.get('type'))))
            path_g.add((previous_node, AGORA.next, b_node))
            previous_node = b_node

        p_node = __get_pattern_node(pattern)
        path_g.add((previous_node, AGORA.byPattern, p_node))

    for i, tp_plan in enumerate(ef_plan):
        paths = tp_plan.get('paths')
        pattern = tp_plan.get('pattern')
        hints = tp_plan.get('hints')
        context = BNode('space_{}'.format(tp_plan.get('context')))
        for path in paths:
            steps = path.get('steps')
            seeds = path.get('seeds')
            if not len(steps) and len(seeds):
                include_path(pattern[2], seeds, steps)
            elif len(steps):
                ty = steps[0].get('type')
                include_path(ty, seeds, steps)

        for t in s_trees:
            plan_graph.set((t, AGORA.length, Literal(tree_lengths.get(t, 0), datatype=XSD.integer)))

        pattern_node = __get_pattern_node(pattern)
        plan_graph.add((context, AGORA.definedBy, pattern_node))
        plan_graph.set((context, RDF.type, AGORA.SearchSpace))
        plan_graph.add((pattern_node, RDF.type, AGORA.TriplePattern))
        (sub, pred, obj) = pattern

        if isinstance(sub, BNode):
            __add_variable(pattern_node, str(sub))

        if isinstance(obj, BNode):
            __add_variable(pattern_node, str(obj), subject=False)
        elif isinstance(obj, Literal):
            node = BNode(str(obj))
            plan_graph.add((pattern_node, AGORA.object, node))
            plan_graph.set((node, RDF.type, AGORA.Literal))
            plan_graph.set((node, AGORA.value, Literal(str(obj), datatype=XSD.string)))
        else:
            plan_graph.add((pattern_node, AGORA.object, obj))

        plan_graph.add((pattern_node, AGORA.predicate, pred))
        if pred == RDF.type:
            if 'check' in hints:
                plan_graph.add((pattern_node, AGORA.checkType, Literal(hints['check'], datatype=XSD.boolean)))

        sub_expected = plan_graph.subjects(predicate=AGORA.expectedType)
        for s in sub_expected:
            expected_types = list(plan_graph.objects(s, AGORA.expectedType))
            for et in expected_types:
                plan_graph.remove((s, AGORA.expectedType, et))
            q_expected_types = [plan_graph.qname(t) for t in expected_types]
            expected_types = [d for d in expected_types if
                              not set.intersection(set(fountain.get_type(plan_graph.qname(d)).get('super')),
                                                   set(q_expected_types))]
            for et in expected_types:
                plan_graph.add((s, AGORA.expectedType, et))

    return plan_graph



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

import requests
import urlparse


class Fountain(object):
    def __init__(self, host):
        self.__fountain_host = host
        self.__types = {}
        self.__properties = {}

    def __send_request(self, path):
        response = requests.get(urlparse.urljoin(self.__fountain_host, path))
        return response.json()

    @property
    def types(self):
        response = self.__send_request('types')
        return response.get('types')

    @property
    def properties(self):
        response = self.__send_request('properties')
        return response.get('properties')

    def get_type_seeds(self, ty):
        response = self.__send_request('seeds/{}'.format(ty))
        return response.get('seeds')

    def get_property(self, prop):
        if prop not in self.__properties:
            self.__properties[prop] = self.__send_request('properties/{}'.format(prop))

        return self.__properties[prop]

    def get_type(self, ty):
        if ty not in self.__types:
            self.__types[ty] = self.__send_request('types/{}'.format(ty))
        return self.__types[ty]

    @property
    def prefixes(self):
        response = self.__send_request('prefixes')
        return response

    def get_property_paths(self, prop):
        response = self.__send_request('paths/{}'.format(prop))
        return response.get("paths")





"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import socket

__author__ = 'Fernando Serena'

from rdflib import RDF
from agora.planner.plan import Plan
from flask import request, make_response, jsonify, render_template, Response
from flask_negotiate import produces
from agora.planner.server import app
from agora.planner.plan.graph import AGORA
import json
import base64
from agora.client.agora import PlanExecutor


@app.route('/plan')
@app.route('/plan/view')
@produces('application/json', 'text/turtle', 'text/html')
def get_plan():
    def __graph_plan(g):
        def __add_node(nid, end=False, shape='roundrectangle', label=None, seed=False):
            node_data = {'data': {'id': base64.b16encode(nid), 'label': nid, 'shape': shape,
                                  'width': max(100, len(nid) * 12)}}
            if label is not None:
                node_data['data']['label'] = str(label)
            if end:
                node_data['classes'] = 'end'
            if seed:
                node_data['classes'] = 'seed'
            if nid in nodes:
                prev_data = nodes[nid]
                if 'classes' in prev_data:
                    node_data['classes'] = prev_data['classes']
            nodes[nid] = node_data

        def __add_edge(source, label, target, end=False):
            eid = base64.b64encode(source + label + target)
            edge_data = {'data': {'id': eid, 'source': nodes[source]['data']['id'], 'label': label + '\n\n',
                                  'target': nodes[target]['data']['id']}}
            if end:
                edge_data['classes'] = 'end'
            if eid in edges:
                prev_data = edges[eid]
                if 'classes' in prev_data:
                    edge_data['classes'] = prev_data['classes']
            edges[eid] = edge_data

        def __check_pattern(parent, link, sources):
            patterns = g.objects(parent, AGORA.byPattern)
            for tp in patterns:
                t_pred = list(g.objects(tp, AGORA.predicate)).pop()
                if t_pred == RDF.type:
                    p_type = g.qname(list(g.objects(tp, AGORA.object)).pop())
                    __add_node(p_type, end=True)
                    if sources is not None:
                        for st in sources:
                            __add_edge(st, link, p_type)
                else:
                    t_pred = g.qname(t_pred)
                    t_obj = list(g.objects(tp, AGORA.object)).pop()
                    if (t_obj, RDF.type, AGORA.Literal) in g:
                        filter_value = list(g.objects(t_obj, AGORA.value)).pop()
                        filter_id = 'n{}'.format(len(nodes))
                        __add_node(filter_id, end=True, shape='ellipse', label='"{}"'.format(filter_value))
                        for st in sources:
                            __add_edge(st, t_pred, filter_id, end=True)
                    else:
                        pred_range = plan.fountain.get_property(t_pred)['range']
                        pred_range = [d for d in pred_range if
                                      not set.intersection(set(plan.fountain.get_type(d).get('super')),
                                                           set(pred_range))]
                        for st in sources:
                            for rt in pred_range:
                                __add_node(rt)
                                __add_edge(st, t_pred, rt, end=True)

        def __follow_next(parent, link=None, sources=None):
            child = list(g.objects(parent, AGORA.next))

            __check_pattern(parent, link, sources)

            for ch in child:
                expected_types = [g.qname(x) for x in g.objects(ch, AGORA.expectedType)]
                for et in expected_types:
                    __add_node(et)
                try:
                    for et in expected_types:
                        if link is not None:
                            for st in sources:
                                __add_edge(st, link, et)
                    on_property = g.qname(list(g.objects(ch, AGORA.onProperty)).pop())
                    last_property = on_property
                    source_types = expected_types
                    __follow_next(ch, last_property, source_types)
                except IndexError:
                    __check_pattern(ch, None, expected_types)

        nodes = {}
        edges = {}
        roots = set([])

        trees = g.subjects(RDF.type, AGORA.SearchTree)

        for tree in trees:
            seed_type = g.qname(list(g.objects(tree, AGORA.fromType)).pop())
            __add_node(seed_type, seed=True)
            roots.add(nodes[seed_type]['data']['id'])
            __follow_next(tree)

        return nodes.values(), edges.values(), list(roots)

    gp_str = request.args.get('gp', '{}')
    plan = Plan(gp_str)

    mimetypes = str(request.accept_mimetypes).split(',')
    if 'application/json' in mimetypes:
        return jsonify(plan.json)

    if 'view' in request.url_rule.rule:
        tps = [tp.strip() for tp in gp_str.replace('"', "'").lstrip('{').rstrip('}').split('.') if tp != '']

        nodes, edges, roots = __graph_plan(plan.graph)
        return render_template('graph.html',
                               nodes=json.dumps(nodes),
                               edges=json.dumps(edges), roots=json.dumps(roots), tps=json.dumps(tps))

    response = make_response(plan.graph.serialize(format='turtle'))
    response.headers['Content-Type'] = 'text/turtle'
    return response


@app.route('/fragment')
def get_fragment():
    def get_triples():
        for prefix, uri in ns:
            yield '@prefix {}: <{}> .\n'.format(prefix, uri)
        yield '\n'
        for (_, s, p, o) in gen:
            yield '{} {} {} .\n'.format(s.n3(graph.namespace_manager), p.n3(graph.namespace_manager),
                                        o.n3(graph.namespace_manager))

    gp_str = request.args.get('gp', '{}')
    plan = Plan(gp_str)
    executor = PlanExecutor(plan.graph)
    gen, ns, graph = executor.get_fragment_generator(queue_wait=1)
    return Response(get_triples(), mimetype='text/n3')


@app.route('/sparql')
def query_fragment():
    q_str = request.args.get('q', None)
    pattern = q_str[q_str.find('{'):]
    plan = Plan(pattern.lstrip().rstrip())
    executor = PlanExecutor(plan.graph)
    fragment = executor.get_fragment()
    result = fragment.query(q_str)
    return jsonify({"result": list(result)})



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

from agora.planner.plan.fountain import Fountain
from agora.planner.server import app
from agora.planner.plan.agp import AgoraGP
from rdflib import RDF, BNode, Literal
from agora.planner.plan.graph import graph_plan
import logging
from rdflib.namespace import FOAF

log = logging.getLogger('agora_planner.plan')


def make_fountain():
    return Fountain(app.config['FOUNTAIN'])


def _stringify_tp(context, (s, p, o)):
    def stringify_elm(elm):
        if isinstance(elm, BNode):
            return elm.n3(context.namespace_manager)
        elif isinstance(elm, Literal):
            return elm.toPython()

        return context.qname(elm)

    return '{} {} {} .'.format(stringify_elm(s), stringify_elm(p), stringify_elm(o))


class Plan(object):
    def __subject_join(self, tp_paths, context, tp1, tp2, **kwargs):
        subject, pr1, o1 = tp1
        _, pr2, o2 = tp2

        log.debug('trying to s-join {} and {}'.format(_stringify_tp(context, tp1), _stringify_tp(context, tp2)))
        if pr2 == RDF.type:
            o2 = context.qname(o2)
            tp2_domain = [o2]
            tp2_domain.extend(self.__fountain.get_type(o2).get('sub'))
        else:
            tp2_domain = self.__fountain.get_property(context.qname(pr2)).get('domain')

        join_paths = tp_paths[tp1][:]

        if pr1 == RDF.type:
            for path in tp_paths[tp1]:
                steps = path.get('steps')
                if len(steps):
                    last_prop = path.get('steps')[-1].get('property')
                    dom_r = self.__fountain.get_property(last_prop).get('range')
                    if len(filter(lambda x: x in tp2_domain, dom_r)):
                        join_paths.remove(path)
                else:
                    join_paths.remove(path)
        elif pr2 == RDF.type:
            for path in tp_paths[tp1]:
                last_type = path.get('steps')[-1].get('type')
                if last_type in tp2_domain:
                    join_paths.remove(path)
        else:
            for path in tp_paths[tp1]:
                if path.get('steps')[-1].get('type') in tp2_domain:
                    join_paths.remove(path)

        return join_paths

    def __subject_object_join(self, tp_paths, context, tp1, tp2, hints=None):
        subject, pr1, o1 = tp1
        _, pr2, o2 = tp2
        log.debug('trying to so-join {} and {}'.format(_stringify_tp(context, tp1), _stringify_tp(context, tp2)))

        pr2 = context.qname(pr2)
        join_paths = tp_paths[tp1][:]

        if pr1 == RDF.type or subject == o2:
            for path in tp_paths[tp1]:
                steps = path.get('steps', [])
                if len(steps):
                    if pr1 == RDF.type:
                        matching_steps = steps[:]
                    else:
                        matching_steps = steps[:-1]
                    for o_path in tp_paths[tp2]:
                        if o_path.get('steps') == matching_steps:
                            join_paths.remove(path)
        elif pr2 == context.qname(RDF.type):
            tp1_range = self.__fountain.get_property(context.qname(pr1)).get('range')
            o2 = context.qname(o2)
            for r_type in tp1_range:
                check_types = self.__fountain.get_type(r_type).get('super')
                check_types.append(r_type)
                if o2 in check_types:
                    join_paths = []
                    break
            if not join_paths and hints is not None:
                hints[tp2]['check'] = hints[tp2].get('check', False) or len(tp1_range) > 1
        else:
            if not subject == o2:
                for path in tp_paths[tp1]:
                    steps = path.get('steps', [])
                    if len(steps):
                        subject_prop = steps[-1].get('property')
                        subject_range = self.__fountain.get_property(subject_prop).get('range')
                        for join_subject in subject_range:
                            if pr2 in self.__fountain.get_type(join_subject).get('properties') and path in join_paths:
                                join_paths.remove(path)
        return join_paths

    def __object_join(self, tp_paths, context, tp1, tp2, **kwargs):
        _, pr1, obj = tp1
        _, pr2, _ = tp2
        log.debug('trying to o-join {} and {}'.format(_stringify_tp(context, tp1), _stringify_tp(context, tp2)))

        tp2_range = self.__fountain.get_property(context.qname(pr2)).get('range')
        tp1_range = self.__fountain.get_property(context.qname(pr1)).get('range')

        if len(filter(lambda x: x in tp1_range, tp2_range)):
            return []

        return tp_paths[tp1]

    def __get_tp_paths(self):
        def __join(f, joins):
            invalid_paths = []
            for (sj, pj, oj) in joins:
                invalid_paths.extend(f(tp_paths, c, (s, pr, o), (sj, pj, oj), hints=tp_hints))
            if len(joins):
                tp_paths[(s, pr, o)] = filter(lambda z: z not in invalid_paths, tp_paths[(s, pr, o)])
            join_paths.extend(invalid_paths)

        tp_paths = {}
        tp_hints = {}
        for c in self.__agp.contexts():
            for (s, pr, o) in c.triples((None, None, None)):
                tp_hints[(s, pr, o)] = {}
                if pr == RDF.type:
                    tp_paths[(s, pr, o)] = self.__fountain.get_property_paths(self.__agp.qname(o))
                else:
                    tp_paths[(s, pr, o)] = self.__fountain.get_property_paths(self.__agp.qname(pr))

            while True:
                join_paths = []

                for (s, pr, o) in c.triples((None, None, None)):
                    if len(tp_paths[(s, pr, o)]):
                        s_join = [(x, pj, y) for (x, pj, y) in c.triples((s, None, None)) if pj != pr]
                        __join(self.__subject_join, s_join)
                        o_join = [(x, pj, y) for (x, pj, y) in c.triples((None, None, o)) if pj != pr]
                        __join(self.__object_join, o_join)
                        so_join = [(x, pj, y) for (x, pj, y) in c.triples((None, None, s))]
                        so_join.extend([(x, pj, y) for (x, pj, y) in c.triples((o, None, None))])
                        __join(self.__subject_object_join, so_join)
                if not join_paths:
                    break

        for (s, pr, o) in tp_hints:
            if pr == RDF.type and 'check' not in tp_hints[(s, pr, o)]:
                tp_hints[(s, pr, o)]['check'] = len(self.__fountain.get_type(self.__agp.qname(o)).get('super')) > 0

        return tp_paths, tp_hints

    def __get_context(self, (s, p, o)):
        return str(list(self.__agp.contexts((s, p, o))).pop().identifier)

    def __init__(self, gp):
        self.__fountain = make_fountain()
        gp = AgoraGP.from_string(gp, self.fountain.prefixes)
        self.__agp = gp.graph
        log.debug('Agora Graph Pattern:\n{}'.format(self.__agp.serialize(format='turtle')))

        paths, hints = self.__get_tp_paths()

        self.__plan = {"plan": [{"context": self.__get_context(tp), "pattern": tp, "paths": path, "hints": hints[tp]}
                                for (tp, path) in paths.items()], "prefixes": gp.prefixes}

        self.__g_plan = graph_plan(self.__plan, self.__fountain)

    @property
    def json(self):
        return self.__plan

    @property
    def graph(self):
        return self.__g_plan

    @property
    def fountain(self):
        return self.__fountain
